<?php

namespace Drupal\config_enforce;

use Drupal\Core\Config\ConfigImporter;
use Drupal\Core\Config\InstallStorage;
use Drupal\Core\Config\StorageComparer;
use Drupal\config\StorageReplaceDataWrapper;
use Drupal\Core\StringTranslation\StringTranslationTrait;

/**
 * Enforces configuration.
 */
class ConfigEnforce {

  use StringTranslationTrait;

  const CONFIG_NAME = 'config_enforce.enforced_configs';

  const CONFIG_ENFORCE_OFF = 0;
  const CONFIG_ENFORCE_NOSUBMIT = 10;
  const CONFIG_ENFORCE_READONLY = 20;

  /**
   * Return whether a given config object is being enforced.
   */
  public static function isEnforced(string $config_name) {
    return !is_null(self::getEnforcedConfig($config_name));
  }

  /**
   * Return the currently configured target module for a given config object.
   */
  public static function getTargetModule(string $config_name) {
    return self::getEnforcedConfig($config_name)['target_module'];
  }

  /**
   * Return the currently configured config directory for a given config object.
   */
  public static function getConfigDirectory(string $config_name) {
    return self::getEnforcedConfig($config_name)['config_directory'];
  }

  /**
   * Return the currently configured path of the config object.
   */
  public static function getConfigFilePath(string $config_name) {
    return self::getEnforcedConfig($config_name)['config_file_path'];
  }

  /**
   * Return the URI of the host form, if one is available.
   */
  public static function getConfigFormUri(string $config_name) {
    return self::getEnforcedConfig($config_name)['config_form_uri'];
  }

  /**
   * Return a list of available enforcement levels.
   */
  public static function getEnforcementLevels() {
    return [
      self::CONFIG_ENFORCE_OFF => t('Allow form & API updates.'),
      self::CONFIG_ENFORCE_NOSUBMIT => t('Allow only API updates.'),
      self::CONFIG_ENFORCE_READONLY => t('Read-only, no updates.'),
    ];
  }

  /**
   * Return the currently configured enforcement level for the config object.
   */
  public static function getEnforcementLevel(string $config_name) {
    return self::getEnforcedConfig($config_name)['enforcement_level'];
  }

  /**
   * Return a human readable enforcement level for the config object.
   */
  public static function getEnforcementLevelLabel(string $config_name) {
    return self::getEnforcementLevels()[self::getEnforcementLevel($config_name)];
  }

  /**
   * Return this module's stored config.
   */
  public static function getConfig($module) {
    return \Drupal::config(self::CONFIG_NAME . '.' . $module);
  }

  /**
   * Return this module's configuration for a given config object.
   */
  public static function getEnforcedConfig(string $config_name) {
    return self::getEnforcedConfigs($config_name);
  }

  /**
   * Return this module's configuration for all config objects, except those
   * from disabled modules.
   */
  public static function getEnforcedConfigs(string $config_name = '') {
    $configs = [];
    foreach (self::getAllModules() as $module => $label) {
      foreach (self::getConfig($module)->get() as $encoded_name => $info) {
        if ($encoded_name == '_core') continue;
        $decoded_name = self::decode($encoded_name);
        // If we were passed a specific config, return just its settings.
        if ($config_name == $decoded_name) return $info;
        $configs[$decoded_name] = $info;
      }
    }
    // If we were passed a specific config, but didn't return its settings above, that config is not enforced.
    if (!empty($config_name)) return null;

    return $configs;
  }

  /**
   * Return a list of modules.
   */
  public static function getAllModules() {
    $modules = [];
    $moduleHandler = \Drupal::moduleHandler();
    foreach ($moduleHandler->getModuleList() as $name => $module) {
      $modules[$name] = $moduleHandler->getName($name);
    }
    asort($modules);
    return $modules;
  }

  /**
   * Consistently encode config names, so that they can be used as array keys when saved.
   *
   * @see decode().
   */
  public static function encode($config_name) {
    return str_replace('.', '__', $config_name);
  }

  /**
   * Consistently decode config names.
   *
   * @see encode().
   */
  public static function decode($encoded_config_name) {
    return str_replace('__', '.', $encoded_config_name);
  }

  /**
   * Re-import all read-only config.
   */
  public static function enforceConfigs() {
    $enforced_configs = self::getEnforcedConfigs();
    foreach ($enforced_configs as $config_name => $setting) {
      if ($setting['enforcement_level'] < self::CONFIG_ENFORCE_READONLY) continue;
      self::importConfig($setting['config_file_path']);
      // @TODO log each import?
    }
  }

  /**
   * Imports a config item from the given filename.
   *
   * @param string $filename
   *   The filename to import from with the Drupal-relative path.
   */
  protected static function importConfig($filename) {
    $contents = @file_get_contents($filename);
    $data = (new InstallStorage())->decode($contents);
    $config_name = basename($filename, '.yml');

    self::handleSpecialCaseConfig($config_name, $data);

    $config_storage = \Drupal::service('config.storage');
    $source_storage = new StorageReplaceDataWrapper($config_storage);
    $source_storage->replaceData($config_name, $data);
    $storage_comparer = new StorageComparer($source_storage, $config_storage);

    $storage_comparer->createChangelist();

    // @TODO log whether config has changed?

    return self::getConfigImporter($storage_comparer)->import();
  }

  /**
   * Return an instantiated ConfigImporter.
   */
  protected static function getConfigImporter($storage_comparer) {
    // @TODO: simplify this when
    // https://www.drupal.org/project/drupal/issues/3123491 is fixed.
    $config_importer = new ConfigImporter(
      $storage_comparer,
      \Drupal::service('event_dispatcher'),
      \Drupal::service('config.manager'),
      \Drupal::service('lock.persistent'),
      \Drupal::service('config.typed'),
      \Drupal::service('module_handler'),
      \Drupal::service('module_installer'),
      \Drupal::service('theme_handler'),
      \Drupal::service('string_translation')
    );

    return $config_importer;
  }

  /**
   * Alter data or otherwise handle config special cases.
   */
  protected static function handleSpecialCaseConfig($config_name, &$data) {
    switch ($config_name) {
      case 'system.site':
        // Site UUID gets validated, so we need to make sure it'll match.
        // @TODO log this?
        $data['uuid'] = \Drupal::config('system.site')->get('uuid');
        break;
      default:
        return;
    }
  }

}
